package qmc

import (
	"bytes"
	"encoding/base64"
	"errors"
	"fmt"
	"io"
	"os"
	"path/filepath"
	"runtime"

	"git.unlock-music.dev/awalol/go-mmkv"
	"github.com/samber/lo"
	"go.uber.org/zap"
	"golang.org/x/exp/slices"
	"golang.org/x/text/unicode/norm"
)

var streamKeyVault mmkv.Vault

// TODO: move to factory
func readKeyFromMMKV(file string, logger *zap.Logger) ([]byte, error) {
	if file == "" {
		return nil, errors.New("file path is required while reading key from mmkv")
	}

	//goland:noinspection GoBoolExpressions
	if runtime.GOOS != "darwin" {
		return nil, errors.New("mmkv vault not supported on this platform")
	}

	if streamKeyVault == nil {
		mmkvDir, err := getRelativeMMKVDir(file)
		if err != nil {
			mmkvDir, err = getDefaultMMKVDir()
			if err != nil {
				return nil, fmt.Errorf("mmkv key valut not found: %w", err)
			}
		}

		mgr, err := mmkv.NewManager(mmkvDir)
		if err != nil {
			return nil, fmt.Errorf("init mmkv manager: %w", err)
		}

		streamKeyVault, err = mgr.OpenVault("MMKVStreamEncryptId")
		if err != nil {
			return nil, fmt.Errorf("open mmkv vault: %w", err)
		}

		logger.Debug("mmkv vault opened", zap.Strings("keys", streamKeyVault.Keys()))
	}

	_, partName := filepath.Split(file)
	partName = normalizeUnicode(partName)

	buf, err := streamKeyVault.GetBytes(file)
	if buf == nil {
		filePaths := streamKeyVault.Keys()
		fileNames := lo.Map(filePaths, func(filePath string, _ int) string {
			_, name := filepath.Split(filePath)
			return normalizeUnicode(name)
		})

		for _, key := range fileNames { // fallback: match filename only
			if key != partName {
				continue
			}
			idx := slices.Index(fileNames, key)
			buf, err = streamKeyVault.GetBytes(filePaths[idx])
			if err != nil {
				logger.Warn("read key from mmkv", zap.String("key", filePaths[idx]), zap.Error(err))
			}
		}

	}

	if len(buf) == 0 {
		return nil, errors.New("key not found in mmkv vault")
	}

	return deriveKey(buf)
}

func OpenMMKV(vaultPath string, vaultKey string, logger *zap.Logger) error {
	filePath, fileName := filepath.Split(vaultPath)
	mgr, err := mmkv.NewManager(filepath.Dir(filePath))
	if err != nil {
		return fmt.Errorf("init mmkv manager: %w", err)
	}

	streamKeyVault, err = mgr.OpenVaultCrypto(fileName, vaultKey)
	if err != nil {
		return fmt.Errorf("open mmkv vault: %w", err)
	}

	logger.Debug("mmkv vault opened", zap.Strings("keys", streamKeyVault.Keys()))
	return nil
}

func readKeyFromMMKVCustom(d *Decoder) ([]byte, error) {
	if streamKeyVault == nil {
		return nil, fmt.Errorf("mmkv vault not loaded")
	}
	// 获取mid即数据库键值
	_, err := d.raw.Seek(-128, io.SeekEnd)
	if err != nil {
		return nil, fmt.Errorf("get mid error: %w", err)
	}
	mid, err := io.ReadAll(io.LimitReader(d.raw, 64))   // 取64字节确保完全取完
	mid = bytes.ReplaceAll(mid, []byte{0x00}, []byte{}) // clean NUL
	mid = bytes.Trim(mid, "\0000")                      // maybe a little stupid

	// 从数据库获取eKey
	eKey, err := streamKeyVault.GetBytes(string(mid))
	if err != nil {
		return nil, fmt.Errorf("get eKey error: %w", err)
	}
	n, err := base64.StdEncoding.Decode(eKey, eKey)
	if err != nil {
		return nil, fmt.Errorf("base64 error: %w", err)
	}
	return deriveKeyV1(eKey[:n])
}

func getRelativeMMKVDir(file string) (string, error) {
	mmkvDir := filepath.Join(filepath.Dir(file), "../mmkv")
	if _, err := os.Stat(mmkvDir); err != nil {
		return "", fmt.Errorf("stat default mmkv dir: %w", err)
	}

	keyFile := filepath.Join(mmkvDir, "MMKVStreamEncryptId")
	if _, err := os.Stat(keyFile); err != nil {
		return "", fmt.Errorf("stat default mmkv file: %w", err)
	}

	return mmkvDir, nil
}

func getDefaultMMKVDir() (string, error) {
	homeDir, err := os.UserHomeDir()
	if err != nil {
		return "", fmt.Errorf("get user home dir: %w", err)
	}

	mmkvDir := filepath.Join(
		homeDir,
		"Library/Containers/com.tencent.QQMusicMac/Data", // todo: make configurable
		"Library/Application Support/QQMusicMac/mmkv",
	)
	if _, err := os.Stat(mmkvDir); err != nil {
		return "", fmt.Errorf("stat default mmkv dir: %w", err)
	}

	keyFile := filepath.Join(mmkvDir, "MMKVStreamEncryptId")
	if _, err := os.Stat(keyFile); err != nil {
		return "", fmt.Errorf("stat default mmkv file: %w", err)
	}

	return mmkvDir, nil
}

// normalizeUnicode normalizes unicode string to NFC.
// since macOS may change some characters in the file name.
// e.g. "ぜ"(e3 81 9c) -> "ぜ"(e3 81 9b e3 82 99)
func normalizeUnicode(str string) string {
	return norm.NFC.String(str)
}
